import { FeedMonitorService } from "./services/feed-monitor-service"
import { ProfileService } from "./services/profile-service"
import { ReadOnlyPostService } from "./services/readonly-post-service"
import { FriendService } from "./services/friend-service"
import { SchemaService } from "./services/schema-service"
import { IdentityService } from "./services/identity-service"
import { ImageService } from "./services/image-service"
import { EventEmitter } from "events"
import { Schema } from "./dto/schema"
import { WalletService } from "./services/wallet-service"
import { PostService } from "./services/post-service"
import { BlogPostService } from "./services/blog-post-service"
import { PageService } from "./services/page-service"
import { SiteSettingsService } from "./services/site-settings-service"

const levelup = require('levelup')
const leveljs = require('level-js')

const IPFS = require('ipfs')
const OrbitDB = require('orbit-db')
const Keystore = require('orbit-db-keystore')

const { providers, ethers } = require('ethers')

const promisify = (inner) =>
  new Promise((resolve, reject) =>
    inner((err, res) => {
      if (err) { reject(err) }
      resolve(res);
    })
  );



namespace Core {

  export var initialized:boolean = false


  /** Etherjs */
  export var provider: any
  export var wallet: any

  /** Orbit db api reference */
  export var orbitDb: any
  export var orbitAccessControl: any  //this is temporary. This will need to be refactored. Remove it and actually create access control

  /** IPFS api client */
  export var ipfs: any

  // export var isElectron: boolean

  export var eventEmitter

  export var walletService: WalletService
  export var walletDao: WalletDao

  export var feedMonitorService: FeedMonitorService
  export var profileService: ProfileService
  export var siteSettingsService: SiteSettingsService

  export var postService:PostService
  export var blogPostService:BlogPostService
  export var readOnlyPostService: ReadOnlyPostService
  export var feedMonitorPostService:ReadOnlyPostService

  export var pageService:PageService
  
  export var friendService: FriendService
  export var schemaService: SchemaService
  export var identityService: IdentityService
  export var imageService: ImageService



  export async function init() {

    Core.schemaService = new SchemaService()
    Core.identityService = new IdentityService()

    console.log('Setting web3 provider')

    if (Core.isElectron()) {
      await configureElectron()
    } else {
      await configureBrowser()
    }



    if (!Core.wallet) {
      throw new Error("No wallet found. Unable to initialize.")
    }


    let walletAddress = await Core.wallet.getAddress()

    console.log(`Initializing Large for wallet: ${walletAddress}`)


    await initOrbit()

    await initSchema(walletAddress)

    initServices()

    //Update user's last know address
    await Core.profileService.updateLastKnownAddress()

    //Connect to online friends
    Core.friendService.connectToFriends(walletAddress)

    //Start feed monitor
    Core.feedMonitorService.start()

    console.log("Initialization complete")

  }

  export async function loadStoresForWallet(walletAddress:string) {

    await Core.blogPostService.loadStoresForWallet(walletAddress)
    await Core.friendService.loadStoreForWallet(walletAddress)
    await Core.pageService.loadStoresForWallet(walletAddress)
    await Core.profileService.loadStoreForWallet(walletAddress)
    await Core.siteSettingsService.loadStoresForWallet(walletAddress)

  }


  export function isElectron() : boolean {
    let userAgent = navigator.userAgent.toLowerCase()
    let result = (userAgent.indexOf(' electron/') > -1) 

    return result
  }




  function initServices() {

    Core.eventEmitter = new EventEmitter()

    //Create services
    // Core.identityService = Core.identityService

    Core.imageService = new ImageService(Core.ipfs)
    Core.profileService = new ProfileService(Core.schemaService)
    Core.siteSettingsService = new SiteSettingsService(Core.schemaService)

    Core.postService = new PostService(Core.schemaService, Core.imageService, Core.profileService)
    Core.blogPostService = new BlogPostService(Core.schemaService, Core.postService)

    Core.readOnlyPostService = new ReadOnlyPostService(Core.schemaService, Core.postService)
    Core.feedMonitorPostService = new ReadOnlyPostService(Core.schemaService, Core.postService)

    Core.pageService = new PageService(Core.schemaService, Core.postService)
    
    Core.friendService = new FriendService(Core.schemaService, Core.profileService)
    Core.feedMonitorService = new FeedMonitorService(Core.schemaService, Core.friendService, Core.feedMonitorPostService)


  }

  async function initSchema(walletAddress: string) {

     //Look up main address
     let mainStore
     try {
       mainStore = await Core.schemaService.getMainStoreByWalletAddress(walletAddress)
     } catch (ex) {
       console.log(ex)
     }
 
 
 
     //If it doesn't exist create it
     if (!mainStore) {
       mainStore = await Core.schemaService.generateMainStore(Core.orbitDb, Core.orbitAccessControl, walletAddress)
     }
 
     //Detect whether or not we already have a schema
     let schema: Schema = await Core.schemaService.getSchema(mainStore, walletAddress)
 
     if (!schema) {
       await Core.schemaService.generateSchema(Core.orbitDb, Core.orbitAccessControl, mainStore, walletAddress)
       schema = await Core.schemaService.getSchema(mainStore, walletAddress)
     }
 
     //Update the schema if it needs it.
     await Core.schemaService.updateSchema(mainStore, schema, walletAddress)

  }

  async function initOrbit() {

    let keystore = new Keystore()

    let identity = await Core.identityService.getIdentity(keystore)

    Core.orbitDb = await OrbitDB.createInstance(Core.ipfs, {
      identity: identity
    })

    Core.orbitAccessControl = Core.identityService.getAccessController(Core.orbitDb)
  }

  async function configureWeb3Browser() {

    //If they're using metamask enable it and use it as a wallet and provider.
    //If not then we're going to use the default provider from ethers
    
    if (window['ethereum']) {

        // Request account access
        await window['ethereum'].enable()

        //@ts-ignore
        window.web3Provider = window.ethereum

        //@ts-ignore
        web3 = new Web3(window.web3Provider)

        //@ts-ignore
        const accounts = await promisify(cb => web3.eth.getAccounts(cb))

        let account = accounts[0]
        window['currentAccount'] = account //TODO: probably something better

        //@ts-ignore
        Core.provider = new providers.Web3Provider(web3.currentProvider)
        Core.wallet = Core.provider.getSigner()

    } else {

      let defaultProviders = ethers.getDefaultProvider("homestead")
      Core.provider = defaultProviders.providers[0]

    }






  }

  function configureWeb3Electron() {


      let defaultProviders = ethers.getDefaultProvider("homestead")
      Core.provider = defaultProviders.providers[0]

  }

  async function configureIpfsElectron() {

    //@ts-ignore
    Core.ipfsHost = remote.getGlobal('ipfsHost')

    Core.ipfs = await IPFS.create({
      EXPERIMENTAL: {
        ipnsPubsub: true
      },
      preload: {
        enabled: false
      },
      relay: {
        enabled: true,
        hop: {
          enabled: true // enable circuit relay HOP (make this node a relay)
        }
      },
      config: {
        Addresses: {
          //@ts-ignore
          Swarm: ['/dns4/ws-star.discovery.libp2p.io/tcp/443/wss/p2p-websocket-star', Core.ipfsHost]
        }
      }
    })

  }

  async function configureIpfsBrowser() {

    Core.ipfs = await IPFS.create({
      EXPERIMENTAL: {
        ipnsPubsub: true
      },
      preload: {
        enabled: false
      },
      relay: {
        enabled: true,
        hop: {
          enabled: true // enable circuit relay HOP (make this node a relay)
        }
      },
      config: {
        Addresses: {
          //@ts-ignore
          Swarm: ['/dns4/ws-star.discovery.libp2p.io/tcp/443/wss/p2p-websocket-star']
        }
      }
    })

  }

  async function configureElectron() {

    //@ts-ignore
    window['remote'] = window.require('electron').remote

    //Wallet DAO
    Core.walletDao = window['remote'].getGlobal('walletDao')

    //Init service
    Core.walletService = new WalletService(Core.walletDao)

    //Web 3
    await configureWeb3Electron()

    //IPFS
    await configureIpfsElectron()

  }

  async function configureBrowser() {

    //Wallet DAO
    Core.walletDao = new WalletDao()

    //Init service
    Core.walletService = new WalletService(Core.walletDao)

    //Web 3
    await configureWeb3Browser()

    //IPFS
    await configureIpfsBrowser()


  }


}


class WalletDao {

  db: any 

  constructor() {
    this.db = levelup(leveljs('large-wallet'))
  }

  async saveWallet(wallet) {
    return this.db.put('wallet', Buffer.from(JSON.stringify(wallet)))
  }

  async loadWallet() {

    try {
      const value = await this.db.get('wallet')

      if (value) {
        return JSON.parse(value.toString())
      }

    } catch(ex) {
      // console.log(ex)
    }



  }
}


export default Core