import { SchemaService } from "./schema-service"
import Core from "../core"
import { PostService } from "./post-service"
import { BlogPost } from "../dto/blog-post"
const sha256 = require('js-sha256')



class BlogPostService {

  public loadedWalletAddress:string 
  public feedStore: any
  public kvStore: any
  

  constructor(
    private schemaService: SchemaService,
    private postService:PostService
  ) { }


  async loadStoresForWallet(walletAddress: string) {

    if (walletAddress == this.loadedWalletAddress) return 

    this.feedStore = await this.schemaService.getBlogPostFeedByWalletAddress(walletAddress)
    this.kvStore = await this.schemaService.getBlogPostKvStoreByWalletAddress(walletAddress)
    
    await this.kvStore.load()

    this.loadedWalletAddress = walletAddress

  }


  async getRecentPosts(offset: number, limit: number, olderThan: string = undefined, newerThan: string = undefined): Promise<BlogPost[]> {

    let results = []

    let loadQuantity = limit + offset

    //Re-open so we can load 
    this.feedStore = await Core.orbitDb.open(this.feedStore.address.toString())
    await this.feedStore.load(loadQuantity)

    results = await this.getPosts(limit, olderThan, newerThan)

    results.reverse()

    return results

  }


  async getPosts(limit: number, olderThan: string = undefined, newerThan: string = undefined): Promise<any> {

    if (!this.feedStore) return []

    let options: any = {
      limit: limit
    }

    if (olderThan) {
      options.lt = olderThan
    }

    if (newerThan) {
      options.gt = newerThan
    }


    let results = await this.feedStore.iterator(options)
      .collect()
      .map((e) => {
        let permalinkKey = e.payload.value
        let cid = this.kvStore.get(permalinkKey)
        return cid
      })


    let posts:BlogPost[] = []

    for (let cid of results) {
      if (cid) posts.push(await BlogPostService.read(cid))  
    }

    return posts
  }



  async create(post: BlogPost): Promise<BlogPost> {

    //Increment counter and set permalink key
    post.permalinkKey = this._generatePermalink(post)

    //Store key in the feed
    post.feedCid = await this.feedStore.add(post.permalinkKey)

    //Store post in IPFS
    let postCid = await this.store(post)
    
    //Associate the postCid with the key
    await this.kvStore.put(post.permalinkKey, postCid.toString())

    return this.translate(post, postCid)

  }


  async update(post:BlogPost) {

    let postCid = await this.store(post)
    await this.kvStore.put(post.permalinkKey, postCid.toString())

    return this.translate(post, postCid)

  }


  /**
   * Store post in IPFS. Returns CID.
   * @param post 
   */
  private async store(post:BlogPost) : Promise<string> {
    let buffer = Buffer.from(JSON.stringify(post))
    return await Core.ipfs.object.put(buffer)
  }

  private translate(post:BlogPost, postCid:string) : BlogPost {
    
    //@ts-ignore
    let returnPost: BlogPost = {}

    Object.assign(returnPost, post)
    returnPost.cid = postCid.toString()

    return returnPost
  }




  static async read(cid: string): Promise<BlogPost> {

    let loaded = await Core.ipfs.object.get(cid)

    let data = loaded.Data ? loaded.Data.toString() : loaded.data.toString()

    let deserialzied = JSON.parse(data)

    //@ts-ignore
    let post: BlogPost = {}
    Object.assign(post, deserialzied)

    post.cid = cid.toString()

    return post
  }


  async readPermalink(permalinkKey:string) {
    return BlogPostService.read(this.kvStore.get(permalinkKey))
  }

  async delete(post: BlogPost): Promise<void> {
    await this.kvStore.del(post.permalinkKey)
    await this.feedStore.remove(post.cid)
  }


  async translatePost(post: BlogPost): Promise<BlogPost> {
    return this.postService.translatePost(post)
  }


  private _generatePermalink(post:BlogPost) : string {
     return sha256(`${post.owner}-${post.dateCreatedMilli}-${post.contentTranslated}-${post.title}-${post.subtitle}`)
  }

  count() {
    return Object.keys(this.kvStore.index).length
  }

  async load(amount) {
    await this.feedStore.load(amount)
  }

  async close() {
    await this.feedStore.close()
    await this.kvStore.close()
  }





}


export {
  BlogPostService
}
