import { ReadOnlyPostService } from "./readonly-post-service"
import { Post } from "../dto/post"
import { FriendService } from "./friend-service"
import { Friend } from "../dto/friend"
import Core from "../core"
import { SchemaService } from "./schema-service"


class FeedMonitorService {

    private _unreadPosts: number = 0

    getUnreadPosts(): number {
        return this._unreadPosts
    }

    constructor(
        private schemaService: SchemaService,
        private friendService: FriendService,
        private readOnlyPostService: ReadOnlyPostService
    ) {}

    async start() {
        await this.getFriendsNewPosts(window['currentAccount'])

        setInterval(async () => { await this.getFriendsNewPosts(window['currentAccount']) }, 60000)
    }

    async getFriendsNewPosts(walletAddress: string) {

        // console.log('start: get new friend posts')

        await this.friendService.loadStoreForWallet(walletAddress)
        let friends: Friend[] = await this.friendService.getAll()

        let newPosts: Post[] = []

        for (let friend of friends) {
            try {
                let friendPosts = await this.getNewPostsFromFriend(friend)
                newPosts = newPosts.concat(friendPosts)
            } catch (ex) {
                console.log(ex)
            }
        }

        
        await this.saveFriendPosts(newPosts, friends, walletAddress)

        // console.log('complete: get new friend posts')

    }

    async getNewPostsFromFriend(friend: Friend): Promise<Post[]> {

        if (!friend.address) return

        let postFeed = await this.schemaService.getReadOnlyPostFeedByWalletAddress(friend.address)

        postFeed = await this.loadUntilCid(postFeed, friend.lastPostFeedCid)


        let posts = await postFeed.iterator({
            limit: -1,
            gt: friend.lastPostFeedCid
        })
            .collect()
            .map((e) => {

                //@ts-ignore
                let post: Post = {}

                Object.assign(post, e.payload.value) //don't want to send back a reference to the actual obj in the store
                post.cid = e.hash

                return post
            })

        return posts

    }

    async loadUntilCid(store, cid: string, limit: number = 100) {

        if (!cid) {

            store = await Core.orbitDb.open(store.address.toString())
            await store.load(limit)

        } else {

            let loopCount = 1
            let lastPostLoaded = false

            let previousLoopLength = 0

            //Load posts until the last post is actually in the feed
            do {

                store = await Core.orbitDb.open(store.address.toString())
                await store.load(loopCount * 5)

                let results = await store.iterator({
                    limit: -1
                })
                    .collect()
                    .map((e) => {
                        if (e.hash == cid) lastPostLoaded = true
                        return e.hash
                    })

                if (!results || results.length == 0 || results.length >= limit || results.length == previousLoopLength) {
                    lastPostLoaded = true
                }

                previousLoopLength = results.length
                loopCount++

            } while (!lastPostLoaded)
        }


        return store


    }

    async saveFriendPosts(posts:Post[], friends:Friend[], walletAddress:string) {

        posts.sort((a, b) => a.dateCreatedMilli > b.dateCreatedMilli ? 1 : -1)

        posts = posts.filter(post => post.owner != undefined)

        if (posts.length > 0) {

            await this.readOnlyPostService.loadMainFeedForWallet(walletAddress)
            await this.readOnlyPostService.load(1) //have to load at least one record or the entire feed gets lost

            for (let post of posts) {

                await this.readOnlyPostService.create(post)

                //Update friend last post seen
                let friend = friends.find(friend => friend.address == post.owner)

                await this.updateFriendLastPost(friend, post)

                //Increment unread post counter
                this._unreadPosts += 1

                Core.eventEmitter.emit('unread-posts-updated', this._unreadPosts)

                console.log(`Post added: ${post.cid}`)
            }

        }
    }

    async updateFriendLastPost(friend: Friend, post: Post) {

        friend.lastPostFeedCid = post.cid
        friend.lastPostMilli = post.dateCreatedMilli

        await this.friendService.put(friend.address, friend)
    }

    async markAllPostsRead() {
        this._unreadPosts = 0
    }

}



export {
    FeedMonitorService
}