import { SchemaService } from "./schema-service"
import { Page } from "../dto/page"
import Core from "../core"
import { PostService } from "./post-service"
import { QuillDeltaToHtmlConverter } from "quill-delta-to-html"
const sha256 = require('js-sha256')
const moment = require('moment')


class PageService {

    public loadedWalletAddress: string
    public kvStore: any

    constructor(
        private schemaService: SchemaService,
        private postService: PostService
    ) { }

    async loadStoresForWallet(walletAddress: string) {

        if (walletAddress == this.loadedWalletAddress) return

        this.kvStore = await this.schemaService.getPageKvStoreByWalletAddress(walletAddress)

        await this.kvStore.load()

        this.loadedWalletAddress = walletAddress

    }

    async getPages(): Promise<Page[]> {

        let pages: Page[] = []

        let keys = Object.keys(this.kvStore.index)

        for (let key of keys) {
            pages.push(await PageService.read(this.kvStore.get(key)))
        }

        return pages
    }


    async create(page: Page): Promise<Page> {

        //Increment counter and set permalink key
        page.permalinkKey = this._generatePermalink(page)

        //Store post in IPFS
        let pageCid = await this.store(page)

        //Associate the postCid with the key
        await this.kvStore.put(page.permalinkKey, pageCid.toString())

        return this.translate(page, pageCid)

    }

    async update(page: Page) {

        let pageCid = await this.store(page)
        await this.kvStore.put(page.permalinkKey, pageCid.toString())

        return this.translate(page, pageCid)

    }



    static async read(cid: string): Promise<Page> {

        let loaded = await Core.ipfs.object.get(cid)

        let data = loaded.Data ? loaded.Data.toString() : loaded.data.toString()

        let deserialzied = JSON.parse(data)

        //@ts-ignore
        let page: Page = {}
        Object.assign(page, deserialzied)

        page.cid = cid.toString()

        return page
    }

    async readPermalink(permalinkKey: string) {
        return PageService.read(this.kvStore.get(permalinkKey))
    }


    async delete(page: Page): Promise<void> {
        await this.kvStore.del(page.permalinkKey)
    }

    async translatePage(page:Page) : Promise<Page> {

        //@ts-ignore
        let translated: Page = {}

        Object.assign(translated, page)

        translated.contentTranslated = await this.postService.translateContent(page)

        //@ts-ignore //TODO: fix this somehow later. view model
        translated.dateCreatedDisplay = moment(page.dateCreatedMilli).fromNow()
        
        return translated
    }


    private _generatePermalink(page: Page): string {
        return sha256(`${page.owner}-${page.dateCreatedMilli}-${page.contentTranslated}-${page.title}`)
    }


    /**
     * Store post in IPFS. Returns CID.
     * @param post 
     */
    private async store(page: Page): Promise<string> {
        let buffer = Buffer.from(JSON.stringify(page))
        return await Core.ipfs.object.put(buffer)
    }


    private translate(page: Page, pageCid: string): Page {

        //@ts-ignore
        let returnPage: Page = {}

        Object.assign(returnPage, page)
        returnPage.cid = pageCid.toString()

        return returnPage
    }




    

    async close() {
        await this.kvStore.close()
    }



}

export {
    PageService
}