import { Schema } from "../dto/schema"
import Core from "../core"
import { Post } from "../dto/post"
import { FeedMonitorService, Friend } from ".."

const OrbitDB = require('orbit-db')
const OrbitDBAddress = require('orbit-db/src/orbit-db-address.js')

const sha256 = require('js-sha256')


class SchemaService {

    private _cachedSchemas = {}


    constructor() {}


    async getSchema(store, walletAddress:string) : Promise<Schema> {

        let schema:Schema

        let results = await store.get(walletAddress)

        if (results && results[0] && results[0].value) {
            schema = results[0].value
        }

        return schema
    }

    async getSchemaByWalletAddress(walletAddress:string) : Promise<Schema> {

        // //Check if we've loaded it already
        let cached = this._cachedSchemas[walletAddress]
        if (cached) return cached


        let mainStore = await this.getMainStoreByWalletAddress(walletAddress)
        let schema:Schema = await this.getSchema(mainStore, walletAddress)

        //Cache it
        this._cachedSchemas[walletAddress] = schema

        if (schema) return schema
        
        return new Promise((resolve, reject) => {
            mainStore.events.on('replicated', async () => {
                console.log(`Replicated main store for ${walletAddress}`)
                let schema:Schema = await this.getSchema(mainStore, walletAddress)

                //Cache it
                this._cachedSchemas[walletAddress] = schema

                resolve(schema)
            })
        })
        
        // throw new Error(`Schema for wallet ${walletAddress} could not be found`)

        return schema
    }

    async getMainStoreByWalletAddress(walletAddress:string) {

        let mainStore
        let mainStoreName = this._getMainStoreNameSeed(walletAddress)

        //get name
        let mainStoreAddress = await Core.orbitDb.determineAddress(mainStoreName, 'docstore', {
            accessController: Core.orbitAccessControl //This might cause issues in the future. Do we need to
        })

        //Try to open it
        mainStore = await Core.orbitDb.open(mainStoreAddress, {
            fetchEntryTimeout: 5000
        })

        await mainStore.load()

        return mainStore

    }

    async getMainFeedByWalletAddress(walletAddress: string) {
        let schema:Schema = await this.getSchemaByWalletAddress(walletAddress)
        return Core.orbitDb.open(schema.mainFeed, {
            fetchEntryTimeout: 5000
        })
    }

    async getProfileStoreByWalletAddress(walletAddress: string) {
        let schema:Schema = await this.getSchemaByWalletAddress(walletAddress)
        return Core.orbitDb.open(schema.profileStore, {
            fetchEntryTimeout: 5000
        })
    }

    async getReadOnlyPostFeedByWalletAddress(walletAddress: string) {
        let schema:Schema = await this.getSchemaByWalletAddress(walletAddress)
        return Core.orbitDb.open(schema.readOnlyPostFeed, {
            fetchEntryTimeout: 5000
        })
    }


    async getBlogPostFeedByWalletAddress(walletAddress: string) {
        let schema:Schema = await this.getSchemaByWalletAddress(walletAddress)
        return Core.orbitDb.open(schema.blogPostFeed, {
            fetchEntryTimeout: 5000
        })
    }

    async getBlogPostKvStoreByWalletAddress(walletAddress: string) {
        let schema:Schema = await this.getSchemaByWalletAddress(walletAddress)
        return Core.orbitDb.open(schema.blogPostKvStore, {
            fetchEntryTimeout: 5000
        })
    }

    async getPageKvStoreByWalletAddress(walletAddress: string) {
        let schema:Schema = await this.getSchemaByWalletAddress(walletAddress)
        return Core.orbitDb.open(schema.pageKvStore, {
            fetchEntryTimeout: 5000
        })
    }


    async getFriendStoreByWalletAddress(walletAddress: string) {
        let schema:Schema = await this.getSchemaByWalletAddress(walletAddress)
        return Core.orbitDb.open(schema.friendStore, {
            fetchEntryTimeout: 5000
        })
    }

    async getSiteSettingsStoreByWalletAddress(walletAddress: string) {
        let schema:Schema = await this.getSchemaByWalletAddress(walletAddress)
        return Core.orbitDb.open(schema.siteSettingsStore, {
            fetchEntryTimeout: 5000
        })
    }

    async reopenStore(store: any) {

        let orbitAddress = store.address.toString()
        await store.close()

        return Core.orbitDb.open(orbitAddress)

    }

    async dropStore(address:string) {
        let store = await Core.orbitDb.open(address)
        await store.drop()
    }


    async getRepliesPostFeed(post:Post, translatedContent: string) {

        let repliesFeedName = this._getRepliesFeedNameSeed(post, translatedContent)

        let repliesFeed = await Core.orbitDb.feed(repliesFeedName, {
            accessController: Core.orbitAccessControl
          })

        return repliesFeed

    }

    async getRepliesPostFeedAddress(post:Post, translatedContent: string) : Promise<string> {
        let feed = await this.getRepliesPostFeed(post, translatedContent)

        let address: string = feed.address.toString()

        return address
    }



    getOrbitAddress(orbitCid:string, walletAddress:string) : string {
        return `/orbitdb/${orbitCid}/mainStore-${walletAddress.toLowerCase()}`
    }



    private _getMainStoreNameSeed(walletAddress:string ): string  {
        return `mainStore-${walletAddress.toLowerCase()}`
    }
    private _getMainFeedNameSeed(walletAddress:string ) : string {
        return `mainFeed-${walletAddress.toLowerCase()}`
    }

    private _getProfileStoreNameSeed(walletAddress:string ): string  {
        return `profile-${walletAddress.toLowerCase()}`
    }

    private _getReadOnlyPostFeedNameSeed(walletAddress:string ) : string {
        return `readonly-post-${walletAddress.toLowerCase()}`
    }

    private _getBlogPostFeedNameSeed(walletAddress:string ) : string {
        return `blog-post-${walletAddress.toLowerCase()}`
    }

    private _getBlogPostKvStoreNameSeed(walletAddress:string ) : string {
        return `blog-post-kv-${walletAddress.toLowerCase()}`
    }

    private _getPageKvStoreNameSeed(walletAddress:string ) : string {
        return `page-kv-${walletAddress.toLowerCase()}`
    }

    private _getFriendStoreNameSeed(walletAddress:string ) : string {
        return `friend-${walletAddress.toLowerCase()}`
    }

    private _getSiteSettingsStoreNameSeed(walletAddress:string ) : string {
        return `site-settings-${walletAddress.toLowerCase()}`
    }



    private _getRepliesFeedNameSeed(post:Post, translatedContent: string) : string {
        let hash = sha256(`${post.owner}-${post.dateCreatedMilli}-${translatedContent}`)
        return `post-${hash}`
    }




    async generateMainStore(orbit, accessController, walletAddress:string) {

        let mainStoreName = this._getMainStoreNameSeed(walletAddress)

        return Core.orbitDb.docstore(mainStoreName, {
            accessController: accessController
        })
    }

    async generateSchema(orbitdb, accessController, mainStore, walletAddress:string) {

        console.log('Generating schema')

        let mainFeed = await this.generateMainFeed(orbitdb, accessController, walletAddress)
        let profileStore = await this.generateProfileStore(orbitdb, accessController, walletAddress)

        let readOnlyPostFeed = await this.generateReadOnlyPostFeed(orbitdb, accessController, walletAddress)

        let blogPostFeed = await this.generateBlogPostFeed(orbitdb, accessController, walletAddress)
        let blogPostKvStore = await this.generateBlogPostKvStore(orbitdb, accessController,walletAddress)

        let pageKvStore = await this.generatePageKvStore(orbitdb, accessController, walletAddress)

        let siteSettingsStore = await this.generateSiteSettingsStore(orbitdb, accessController, walletAddress)
        let friendStore = await this.generateFriendStore(orbitdb, accessController, walletAddress)

        let schema:Schema = {
          
          mainFeed: mainFeed.address.toString(),
          profileStore: profileStore.address.toString(),
          readOnlyPostFeed: readOnlyPostFeed.address.toString(),

          blogPostFeed: blogPostFeed.address.toString(),
          blogPostKvStore: blogPostKvStore.address.toString(),

          pageKvStore: pageKvStore.address.toString(),
          
          siteSettingsStore: siteSettingsStore.address.toString(),
          friendStore: friendStore.address.toString()
        }

        await mainStore.put({
          _id: walletAddress,
          value: schema
        })

        console.log('Inserted schema into mainStore')

    }


    async updateSchema(mainStore, schema:Schema, walletAddress:string) {

        //Make sure schema has all fields
        let schemaUpdated:boolean = false

        if (!schema.mainFeed) {
            let mainFeed = await this.generateMainFeed(Core.orbitDb, Core.orbitAccessControl, walletAddress)
            schema.mainFeed = mainFeed.address.toString()
            schemaUpdated = true
        }

        if (!schema.profileStore) {
            let profileStore = await this.generateProfileStore(Core.orbitDb, Core.orbitAccessControl, walletAddress)
            schema.profileStore = profileStore.address.toString()
            schemaUpdated = true
        }

        if (!schema.readOnlyPostFeed) {
            let readOnlyPostFeed = await this.generateReadOnlyPostFeed(Core.orbitDb, Core.orbitAccessControl, walletAddress)
            schema.readOnlyPostFeed = readOnlyPostFeed.address.toString()
            schemaUpdated = true
        }

        if (!schema.blogPostFeed) {
            let blogPostFeed = await this.generateBlogPostFeed(Core.orbitDb, Core.orbitAccessControl, walletAddress)
            schema.blogPostFeed = blogPostFeed.address.toString()
            schemaUpdated = true
        }

        if (!schema.blogPostKvStore) {
            let blogPostKvStore = await this.generateBlogPostKvStore(Core.orbitDb, Core.orbitAccessControl, walletAddress)
            schema.blogPostKvStore = blogPostKvStore.address.toString()
            schemaUpdated = true
        }

        if (!schema.pageKvStore) {
            let pageKvStore = await this.generatePageKvStore(Core.orbitDb, Core.orbitAccessControl, walletAddress)
            schema.pageKvStore = pageKvStore.address.toString()
            schemaUpdated = true
        }


        if (!schema.siteSettingsStore) {
            let siteSettingsStore = await this.generateSiteSettingsStore(Core.orbitDb, Core.orbitAccessControl, walletAddress)
            schema.siteSettingsStore = siteSettingsStore.address.toString()
            schemaUpdated = true
        }

        if (!schema.friendStore) {
            let friendStore = await this.generateFriendStore(Core.orbitDb, Core.orbitAccessControl, walletAddress)
            schema.friendStore = friendStore.address.toString()
            schemaUpdated = true
        }


        if (schemaUpdated) {

            console.log("Updating schema")

            await mainStore.put({
                _id: walletAddress,
                value: schema
            })
        }

    }


    async generateMainFeed(orbitdb, accessController, walletAddress:string) {

        console.log("Generating main feed")

        let mainFeedName = this._getMainFeedNameSeed(walletAddress)

        return orbitdb.feed(mainFeedName, {
          create: true,
          accessController: accessController
        })

    }



    async generateProfileStore(orbitdb, accessController, walletAddress:string) {

        console.log("Generating profile store")

        //Create profile store
        let profileStoreName = this._getProfileStoreNameSeed(walletAddress)

        return orbitdb.docstore(profileStoreName, {
          create: true,
          accessController: accessController
        })

    }


    async generateReadOnlyPostFeed(orbitdb, accessController, walletAddress:string) {

        console.log("Generating read only post feed")

        let readOnlyPostFeedName = this._getReadOnlyPostFeedNameSeed(walletAddress)

        return orbitdb.feed(readOnlyPostFeedName, {
          create: true,
          accessController: accessController
        })

    }

    async generateBlogPostFeed(orbitdb, accessController, walletAddress:string) {

        console.log("Generating blog post feed")

        let blogPostFeedName = this._getBlogPostFeedNameSeed(walletAddress)

        return orbitdb.feed(blogPostFeedName, {
          create: true,
          accessController: accessController
        })

    }

    async generateBlogPostKvStore(orbitdb, accessController, walletAddress:string) {

        console.log("Generating blog post kv store")

        let blogPostKvStoreName = this._getBlogPostKvStoreNameSeed(walletAddress)

        return orbitdb.kvstore(blogPostKvStoreName, {
          create: true,
          accessController: accessController
        })

    }


    async generatePageKvStore(orbitdb, accessController, walletAddress:string) {

        console.log("Generating page kv store")

        let pageKvStoreName = this._getPageKvStoreNameSeed(walletAddress)

        return orbitdb.kvstore(pageKvStoreName, {
          create: true,
          accessController: accessController
        })

    }


    async generateFriendStore(orbitdb, accessController, walletAddress:string) {

        console.log("Generating friend store")

        let friendStoreName = this._getFriendStoreNameSeed(walletAddress)

        return orbitdb.kvstore(friendStoreName, {
            create: true,
            accessController: accessController
        })

    }


    async generateSiteSettingsStore(orbitdb, accessController, walletAddress:string) {

        console.log("Generating site settings store")

        let siteSettingsStoreName = this._getSiteSettingsStoreNameSeed(walletAddress)

        return orbitdb.kvstore(siteSettingsStoreName, {
            create: true,
            accessController: accessController
        })

    }


    async openAddress(address:any) {

        let orbitAddress = new OrbitDBAddress(address.root, address.path)

        let parsedAddress = orbitAddress.toString()
        return Core.orbitDb.open(parsedAddress)
    }


}



export {
    SchemaService
}
