import assert = require('assert')
import { BlogPostService } from "../src/services/blog-post-service"
import Core from "../src/core"
import { SchemaService } from "../src/services/schema-service"

const moment = require('moment')
const OrbitDB = require('orbit-db')

import { getIPFS } from './ipfs'
import { PostService } from "../src/services/post-service"
import { ImageService, ProfileService } from "../src"
import { BlogPost } from "../src/dto/blog-post"


//@ts-ignore
describe('BlogPostService', async (accounts) => {

    let service: BlogPostService
    let mainStore
    let address: string

    let ipfs

    //@ts-ignore
    before("", async () => {

        ipfs = await getIPFS()

        address = Math.random().toString()

        const orbitdb = await OrbitDB.createInstance(ipfs, {
            directory: "./test/orbitdb/" + Math.random().toString()
        })

        Core.ipfs = ipfs
        Core.orbitDb = orbitdb
        Core.schemaService = new SchemaService()
        Core.imageService = new ImageService(Core.ipfs)
        Core.profileService = new ProfileService(Core.schemaService)
        Core.postService = new PostService(Core.schemaService, Core.imageService, Core.profileService)

        mainStore = await Core.schemaService.getMainStoreByWalletAddress(address)
        await mainStore.load()

        await Core.schemaService.generateSchema(Core.orbitDb, Core.orbitAccessControl, mainStore, address)

        service = new BlogPostService(Core.schemaService, Core.postService)

        await service.loadStoresForWallet(address)


    })

    //@ts-ignore
    after("After", async () => {
        // await ipfs.stop()
    })


    let permalink1 

    //@ts-ignore
    it("should create & get", async () => {

        //Arrange
        let post: BlogPost = {
            content: "Actual content",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        }

        //Act
        post = await service.create(post)

        //Assert
        assert.notEqual(post.cid, undefined)

        //Read via cid
        let fetched: BlogPost = await BlogPostService.read(post.cid)

        assert.equal(fetched.content, "Actual content")
        assert.equal(fetched.cid, post.cid)

        permalink1 = fetched.permalinkKey

        //Read via permalinkKey
        fetched = await service.readPermalink(permalink1)

        assert.equal(fetched.content, "Actual content")
        assert.equal(fetched.cid, post.cid)
        assert.equal(fetched.permalinkKey, permalink1)

    })

    //@ts-ignore
    it("should update a post", async () => {

        //Arrange
        let post: BlogPost = await service.readPermalink(permalink1)
        post.content = "Updated content"


        //Act
        let updated: BlogPost = await service.update(post)


        //Assert

        //Read via cid
        let fetched: BlogPost = await BlogPostService.read(updated.cid)

        assert.equal(fetched.content, "Updated content")
        assert.equal(fetched.cid, updated.cid)
        assert.equal(fetched.permalinkKey, permalink1)


        //Read via permalinkKey
        fetched = await service.readPermalink(permalink1)

        assert.equal(fetched.content, "Updated content")
        assert.equal(fetched.cid, updated.cid)
        assert.equal(fetched.permalinkKey, permalink1)


        //Read from feed
        let it = await service.getRecentPosts(0, permalink1)

        assert.equal(it[0].content, "Updated content")
        assert.equal(it[0].cid, updated.cid)
        assert.equal(it[0].permalinkKey, permalink1)

    })

    let permalink2
    let permalink3
    let permalink4

    //@ts-ignore
    it("should create multiple posts and read back in order", async () => {

        //Arrange
        let post2 = await service.create({
            content: "1",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })

        let post3 = await service.create({
            content: "2",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })

        let post4 = await service.create({
            content: "3",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })

        permalink2 = post2.permalinkKey
        permalink3 = post3.permalinkKey
        permalink4 = post4.permalinkKey


        //Act
        let it = await service.getRecentPosts(0, 3)

        //assert
        assert.equal(it.length, 3)
        assert.equal(it[0].content, "3")
        assert.equal(it[1].content, "2")
        assert.equal(it[2].content, "1")
    })

    //@ts-ignore
    it("should update those posts and still read them back in order", async () => {

        //Arrange
        let post1: BlogPost = await service.readPermalink(permalink2)
        post1.content = "1-update"

        let post2: BlogPost = await service.readPermalink(permalink3)
        post2.content = "2-update"

        let post3: BlogPost = await service.readPermalink(permalink4)
        post3.content = "3-update"


        //Act
        await service.update(post1)
        await service.update(post2)
        await service.update(post3)


        //Act
        let it = await service.getRecentPosts(0, 3)

        //assert
        assert.equal(it.length, 3)
        assert.equal(it[0].content, "3-update")
        assert.equal(it[1].content, "2-update")
        assert.equal(it[2].content, "1-update")


    })


    //@ts-ignore
    it("should create multiple posts and read back just the last part of the list", async () => {

        //Arrange
        await service.create({
            content: "4",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })

        await service.create({
            content: "5",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })


        await service.create({
            content: "6",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })


        //Act
        let it = await service.getRecentPosts(0, 3)


        //assert
        assert.equal(it.length, 3)
        assert.equal(it[0].content, "6")
        assert.equal(it[1].content, "5")
        assert.equal(it[2].content, "4")
    })


    //@ts-ignore
    it("should create multiple posts and skip a few of them", async () => {

        //Arrange
        let post = await service.create({
            content: "7",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })


        await service.create({
            content: "8",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })

        await service.create({
            content: "9",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })


        //Act
        let it = await service.getRecentPosts(7, 3, post.feedCid)


        //assert
        assert.equal(it.length, 3)
        assert.equal(it[0].content, "6")
        assert.equal(it[1].content, "5")
        assert.equal(it[2].content, "4")
    })



    //@ts-ignore
    it("should load a database with lots of records and page through them", async () => {

        //Arrange
        for (var i = 0; i < 100; i++) {
            await service.create({
                content: (i + 10).toString(),
                owner: address.toString(),
                dateCreatedMilli: moment().utc().valueOf()
            })
        }

        await service.close()

        await service.loadStoresForWallet(address)

        //Get a page of 3
        let it = await service.getRecentPosts(0, 3)


        //assert
        assert.equal(it.length, 3)
        assert.equal(it[0].content, "109")
        assert.equal(it[1].content, "108")
        assert.equal(it[2].content, "107")

        await service.loadStoresForWallet(address)
        it = await service.getRecentPosts(3, 3, it[2].feedCid)

        assert.equal(it.length, 3)
        assert.equal(it[0].content, "106")
        assert.equal(it[1].content, "105")
        assert.equal(it[2].content, "104")


        await service.loadStoresForWallet(address)
        it = await service.getRecentPosts(6, 3, it[2].feedCid)

        assert.equal(it.length, 3)
        assert.equal(it[0].content, "103")
        assert.equal(it[1].content, "102")
        assert.equal(it[2].content, "101")


        await service.loadStoresForWallet(address)
        it = await service.getRecentPosts(9, 3, it[2].feedCid)

        assert.equal(it.length, 3)
        assert.equal(it[0].content, "100")
        assert.equal(it[1].content, "99")
        assert.equal(it[2].content, "98")

        await service.loadStoresForWallet(address)
        it = await service.getRecentPosts(12, 3, it[2].feedCid)

        assert.equal(it.length, 3)
        assert.equal(it[0].content, "97")
        assert.equal(it[1].content, "96")
        assert.equal(it[2].content, "95")

    })


    //@ts-ignore 
    it("should should page properly no matter what limit we pass", async () => {

        let mainStore2 = await Core.schemaService.getMainStoreByWalletAddress("test-2")
        await mainStore2.load()
        await Core.schemaService.generateSchema(Core.orbitDb, Core.orbitAccessControl, mainStore2, "test-2")

        await service.loadStoresForWallet("test-2")

        //Arrange
        for (var i = 0; i < 10; i++) {
            await service.create({
                content: (i + 10).toString(),
                owner: address.toString(),
                dateCreatedMilli: moment().utc().valueOf()
            })
        }

        await service.close()
        await service.loadStoresForWallet("test-2")
        let less = await service.getRecentPosts(0, 8)
        assert.equal(less.length, 8)


        //All records
        await service.close()
        await service.loadStoresForWallet("test-2")
        let all = await service.getRecentPosts(0, 10)
        assert.equal(all.length, 10)

        await service.close()
        await service.loadStoresForWallet("test-2")
        let more = await service.getRecentPosts(0, 11)
        assert.equal(more.length, 10)


        await service.close()
        await service.loadStoresForWallet("test-2")
        let more2 = await service.getRecentPosts(0, 12)
        assert.equal(more2.length, 10)

        await service.close()
        await service.loadStoresForWallet("test-2")
        let more3 = await service.getRecentPosts(0, 14)
        assert.equal(more3.length, 10)


        await service.close()
        await service.loadStoresForWallet("test-2")
        let more4 = await service.getRecentPosts(0, 14)
        assert.equal(more4.length, 10)


    })







})
