import { Post } from "../src/dto/post"
import assert = require('assert')
import { ReadOnlyPostService } from "../src/services/readonly-post-service"
import Core from "../src/core";
import { SchemaService } from "../src/services/schema-service";
import { ProfileService } from "../src/services/profile-service";

const moment = require('moment')
const OrbitDB = require('orbit-db')

import { getIPFS } from './ipfs'
import { PostService } from "../src/services/post-service";




//@ts-ignore
describe('ReadOnlyPostService', async (accounts) => {

    let service: ReadOnlyPostService
    let mainStore
    let address: string

    let ipfs

    //@ts-ignore
    before("", async () => {

        ipfs = await getIPFS()

        address = Math.random().toString()

        const orbitdb = await OrbitDB.createInstance(ipfs, {
            directory: "./test/orbitdb/" + Math.random().toString()
        })

        Core.ipfs = ipfs
        Core.orbitDb = orbitdb
        Core.schemaService = new SchemaService()
        Core.profileService = new ProfileService(Core.schemaService)
        Core.postService = new PostService(Core.schemaService, Core.imageService, Core.profileService)

        mainStore = await Core.schemaService.getMainStoreByWalletAddress(address)
        await mainStore.load()

        await Core.schemaService.generateSchema(Core.orbitDb, Core.orbitAccessControl, mainStore, address)

        service = new ReadOnlyPostService(Core.schemaService, Core.postService)

        await service.loadPostFeedForWallet(address)


    })

    //@ts-ignore
    after("After", async () => {
        // await ipfs.stop()
    })

    //@ts-ignore
    it("should create & get", async () => {

        //Arrange
        let post: Post = {
            content: "Actual content",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        }

        //Act
        post = await service.create(post)

        //Assert
        assert.notEqual(post.cid, undefined)


        let fetched: Post = await ReadOnlyPostService.read(post.cid)


        assert.equal(fetched.content, "Actual content")
        assert.equal(fetched.cid, post.cid)
    })

    //@ts-ignore
    it("should create multiple posts and read back in order", async () => {

        //Arrange
        await service.create({
            content: "1",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })

        await service.create({
            content: "2",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })

        await service.create({
            content: "3",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })


        //Act
        let it = await service.getRecentPosts(0, 3)

        //assert
        assert.equal(it.length, 3)
        assert.equal(it[0].content, "3")
        assert.equal(it[1].content, "2")
        assert.equal(it[2].content, "1")
    })


    //@ts-ignore
    it("should create multiple posts and read back just the last part of the list", async () => {

        //Arrange
        await service.create({
            content: "4",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })

        await service.create({
            content: "5",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })


        await service.create({
            content: "6",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })


        //Act
        let it = await service.getRecentPosts(0, 3)


        //assert
        assert.equal(it.length, 3)
        assert.equal(it[0].content, "6")
        assert.equal(it[1].content, "5")
        assert.equal(it[2].content, "4")
    })


    //@ts-ignore
    it("should create multiple posts and skip a few of them", async () => {

        //Arrange
        let post = await service.create({
            content: "7",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })


        await service.create({
            content: "8",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })

        await service.create({
            content: "9",
            owner: address.toString(),
            dateCreatedMilli: moment().utc().valueOf()
        })


        //Act
        let it = await service.getRecentPosts(7, 3, post.cid)


        //assert
        assert.equal(it.length, 3)
        assert.equal(it[0].content, "6")
        assert.equal(it[1].content, "5")
        assert.equal(it[2].content, "4")
    })



    //@ts-ignore
    it("should load a database with lots of records and page through them", async () => {

        //Arrange
        for (var i = 0; i < 100; i++) {
            await service.create({
                content: (i + 10).toString(),
                owner: address.toString(),
                dateCreatedMilli: moment().utc().valueOf()
            })
        }

        await service.close()

        await service.loadPostFeedForWallet(address)

        //Get a page of 3
        let it = await service.getRecentPosts(0, 3)


        //assert
        assert.equal(it.length, 3)
        assert.equal(it[0].content, "109")
        assert.equal(it[1].content, "108")
        assert.equal(it[2].content, "107")

        await service.loadPostFeedForWallet(address)
        it = await service.getRecentPosts(3, 3, it[2].cid)

        assert.equal(it.length, 3)
        assert.equal(it[0].content, "106")
        assert.equal(it[1].content, "105")
        assert.equal(it[2].content, "104")


        await service.loadPostFeedForWallet(address)
        it = await service.getRecentPosts(6, 3, it[2].cid)

        assert.equal(it.length, 3)
        assert.equal(it[0].content, "103")
        assert.equal(it[1].content, "102")
        assert.equal(it[2].content, "101")


        await service.loadPostFeedForWallet(address)
        it = await service.getRecentPosts(9, 3, it[2].cid)

        assert.equal(it.length, 3)
        assert.equal(it[0].content, "100")
        assert.equal(it[1].content, "99")
        assert.equal(it[2].content, "98")

        await service.loadPostFeedForWallet(address)
        it = await service.getRecentPosts(12, 3, it[2].cid)

        assert.equal(it.length, 3)
        assert.equal(it[0].content, "97")
        assert.equal(it[1].content, "96")
        assert.equal(it[2].content, "95")

    })


    //@ts-ignore 
    it("should should page properly no matter what limit we pass", async () => {

        let mainStore2 = await Core.schemaService.getMainStoreByWalletAddress("test-2")
        await mainStore2.load()
        await Core.schemaService.generateSchema(Core.orbitDb, Core.orbitAccessControl, mainStore2, "test-2")

        await service.loadPostFeedForWallet("test-2")

        //Arrange
        for (var i = 0; i < 10; i++) {
            await service.create({
                content: (i + 10).toString(),
                owner: address.toString(),
                dateCreatedMilli: moment().utc().valueOf()
            })
        }

        await service.close()
        await service.loadPostFeedForWallet("test-2")
        let less = await service.getRecentPosts(0, 8)
        assert.equal(less.length, 8)


        //All records
        await service.close()
        await service.loadPostFeedForWallet("test-2")
        let all = await service.getRecentPosts(0, 10)
        assert.equal(all.length, 10)

        await service.close()
        await service.loadPostFeedForWallet("test-2")
        let more = await service.getRecentPosts(0, 11)
        assert.equal(more.length, 10)


        await service.close()
        await service.loadPostFeedForWallet("test-2")
        let more2 = await service.getRecentPosts(0, 12)
        assert.equal(more2.length, 10)

        await service.close()
        await service.loadPostFeedForWallet("test-2")
        let more3 = await service.getRecentPosts(0, 14)
        assert.equal(more3.length, 10)


        await service.close()
        await service.loadPostFeedForWallet("test-2")
        let more4 = await service.getRecentPosts(0, 14)
        assert.equal(more4.length, 10)


    })


})
